# AutoBenchmark

Determine the speed of various AceSorting functions, for various array sizes.

**Version**: AceSorting v0.3

**DO NOT EDIT**: This file was auto-generated using `make README.md`.

## Dependencies

This program depends on the following libraries:

* [AceCommon](https://github.com/bxparks/AceCommon)

## How to Generate

This requires the [AUniter](https://github.com/bxparks/AUniter) script
to execute the Arduino IDE programmatically.

The `Makefile` has rules to generate the `*.txt` results file for several
microcontrollers that I usually support, but the `$ make benchmarks` command
does not work very well because the USB port of the microcontroller is a
dynamically changing parameter. I created a semi-automated way of collect the
`*.txt` files:

1. Connect the microcontroller to the serial port. I usually do this through a
USB hub with individually controlled switch.
2. Type `$ auniter ports` to determine its `/dev/ttyXXX` port number (e.g.
`/dev/ttyUSB0` or `/dev/ttyACM0`).
3. If the port is `USB0` or `ACM0`, type `$ make nano.txt`, etc.
4. Switch off the old microontroller.
5. Go to Step 1 and repeat for each microcontroller.

The `generate_table.awk` program reads one of `*.txt` files and prints out an
ASCII table that can be directly embedded into this README.md file. For example
the following command produces the table in the Nano section below:

```
$ ./generate_table.awk < nano.txt
```

Fortunately, we no longer need to run `generate_table.awk` for each `*.txt`
file. The process has been automated using the `generate_readme.py` script which
will be invoked by the following command:
```
$ make README.md
```

## CPU Time Changes

**v0.1**

* Initial results.
* The C-library `qsort()` is far slower than the C++ version because it uses
  a callback to a comparison function through a function pointer.

**v0.3**

* No performance change after rerouting 2-argument sorting functions into
  the 3-argument versions using a default lambda expression.

## Results

The following results show the runtime of each sorting function in milliseconds,
when sorting different array sizes.

### Arduino Nano

* 16MHz ATmega328P
* Arduino IDE 1.8.13
* Arduino AVR Boards 1.8.3
* `micros()` has a resolution of 4 microseconds

```
+---------------------+------------------------+---------+---------+---------+
|            \      N |    10 |    30 |    100 |     300 |    1000 |    3000 |
| Function    \       |       |       |        |         |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| bubbleSort()        | 0.099 | 1.044 | 12.305 | 118.589 |         |         |
| insertionSort()     | 0.049 | 0.251 |  2.463 |  21.288 |         |         |
| selectionSort()     | 0.087 | 0.553 |  5.365 |  46.275 |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| shellSortClassic()  | 0.074 | 0.306 |  1.711 |   6.868 |         |         |
| shellSortKnuth()    | 0.100 | 0.325 |  1.431 |   5.683 |         |         |
| shellSortTokuda()   | 0.075 | 0.336 |  1.616 |   6.555 |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| combSort13()        | 0.167 | 0.534 |  2.215 |   8.260 |         |         |
| combSort13m()       | 0.164 | 0.557 |  2.233 |   8.176 |         |         |
| combSort133()       | 0.084 | 0.400 |  1.968 |   7.730 |         |         |
| combSort133m()      | 0.087 | 0.419 |  1.979 |   7.805 |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| quickSortMiddle()   | 0.097 | 0.373 |  1.582 |   5.536 |         |         |
| quickSortMedian()   | 0.117 | 0.423 |  1.717 |   5.905 |         |         |
| quickSortMdnSwppd() | 0.092 | 0.338 |  1.396 |   4.895 |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| qsort()             | 0.208 | 0.827 |  3.642 |  13.068 |         |         |
+---------------------+-------+-------+--------+---------+---------+---------+

```

### SparkFun Pro Micro

* 16 MHz ATmega32U4
* Arduino IDE 1.8.13
* SparkFun AVR Boards 1.1.13
* `micros()` has a resolution of 4 microseconds

```
+---------------------+------------------------+---------+---------+---------+
|            \      N |    10 |    30 |    100 |     300 |    1000 |    3000 |
| Function    \       |       |       |        |         |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| bubbleSort()        | 0.091 | 1.136 | 12.908 | 120.933 |         |         |
| insertionSort()     | 0.048 | 0.271 |  2.652 |  22.188 |         |         |
| selectionSort()     | 0.084 | 0.556 |  5.391 |  46.513 |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| shellSortClassic()  | 0.075 | 0.314 |  1.730 |   6.950 |         |         |
| shellSortKnuth()    | 0.100 | 0.330 |  1.455 |   5.634 |         |         |
| shellSortTokuda()   | 0.075 | 0.337 |  1.645 |   6.555 |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| combSort13()        | 0.162 | 0.538 |  2.234 |   8.211 |         |         |
| combSort13m()       | 0.164 | 0.559 |  2.266 |   8.232 |         |         |
| combSort133()       | 0.088 | 0.390 |  1.996 |   7.672 |         |         |
| combSort133m()      | 0.087 | 0.424 |  1.993 |   7.763 |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| quickSortMiddle()   | 0.098 | 0.372 |  1.617 |   5.667 |         |         |
| quickSortMedian()   | 0.116 | 0.432 |  1.720 |   5.956 |         |         |
| quickSortMdnSwppd() | 0.090 | 0.345 |  1.391 |   4.914 |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| qsort()             | 0.202 | 0.900 |  3.674 |  13.046 |         |         |
+---------------------+-------+-------+--------+---------+---------+---------+

```

### SAMD21 M0 Mini

* 48 MHz ARM Cortex-M0+
* Arduino IDE 1.8.13
* SparkFun SAMD Core 1.8.3

```
+---------------------+------------------------+---------+---------+---------+
|            \      N |    10 |    30 |    100 |     300 |    1000 |    3000 |
| Function    \       |       |       |        |         |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| bubbleSort()        | 0.031 | 0.282 |  3.405 |  32.272 | 363.409 |         |
| insertionSort()     | 0.016 | 0.088 |  0.941 |   8.144 |  89.678 |         |
| selectionSort()     | 0.024 | 0.162 |  1.629 |  14.306 | 157.639 |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| shellSortClassic()  | 0.016 | 0.064 |  0.349 |   1.408 |   5.912 |  22.251 |
| shellSortKnuth()    | 0.019 | 0.065 |  0.298 |   1.154 |   5.074 |  19.321 |
| shellSortTokuda()   | 0.019 | 0.081 |  0.390 |   1.568 |   6.690 |  24.190 |
|---------------------+-------+-------+--------+---------+---------+---------|
| combSort13()        | 0.035 | 0.121 |  0.532 |   1.934 |   8.697 |  31.638 |
| combSort13m()       | 0.036 | 0.128 |  0.534 |   1.931 |   8.274 |  29.531 |
| combSort133()       | 0.021 | 0.090 |  0.452 |   1.825 |   8.259 |  28.640 |
| combSort133m()      | 0.021 | 0.100 |  0.470 |   1.822 |   7.949 |  28.592 |
|---------------------+-------+-------+--------+---------+---------+---------|
| quickSortMiddle()   | 0.022 | 0.077 |  0.312 |   1.097 |   4.246 |  14.192 |
| quickSortMedian()   | 0.026 | 0.088 |  0.338 |   1.158 |   4.387 |  14.495 |
| quickSortMdnSwppd() | 0.020 | 0.069 |  0.281 |   0.980 |   3.720 |  12.486 |
|---------------------+-------+-------+--------+---------+---------+---------|
| qsort()             | 0.052 | 0.204 |  0.873 |   3.088 |  12.102 |  41.217 |
+---------------------+-------+-------+--------+---------+---------+---------+

```

### STM32

* STM32 "Blue Pill", STM32F103C8, 72 MHz ARM Cortex-M3
* Arduino IDE 1.8.13
* STM32duino 2.0.0

```
+---------------------+------------------------+---------+---------+---------+
|            \      N |    10 |    30 |    100 |     300 |    1000 |    3000 |
| Function    \       |       |       |        |         |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| bubbleSort()        | 0.022 | 0.169 |  2.156 |  20.662 | 230.118 |         |
| insertionSort()     | 0.009 | 0.054 |  0.488 |   4.278 |  45.077 |         |
| selectionSort()     | 0.016 | 0.102 |  1.023 |   8.931 |  98.184 |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| shellSortClassic()  | 0.014 | 0.057 |  0.305 |   1.243 |   5.157 |  19.747 |
| shellSortKnuth()    | 0.013 | 0.057 |  0.276 |   1.131 |   5.058 |  19.403 |
| shellSortTokuda()   | 0.012 | 0.048 |  0.226 |   0.905 |   3.888 |  14.123 |
|---------------------+-------+-------+--------+---------+---------+---------|
| combSort13()        | 0.017 | 0.080 |  0.396 |   1.601 |   7.173 |  26.248 |
| combSort13m()       | 0.018 | 0.084 |  0.416 |   1.585 |   6.998 |  25.575 |
| combSort133()       | 0.016 | 0.073 |  0.378 |   1.481 |   6.775 |  22.826 |
| combSort133m()      | 0.017 | 0.080 |  0.382 |   1.496 |   6.436 |  23.289 |
|---------------------+-------+-------+--------+---------+---------+---------|
| quickSortMiddle()   | 0.017 | 0.064 |  0.260 |   0.941 |   3.559 |  12.330 |
| quickSortMedian()   | 0.020 | 0.070 |  0.280 |   0.994 |   3.769 |  12.632 |
| quickSortMdnSwppd() | 0.015 | 0.056 |  0.236 |   0.855 |   3.318 |  11.345 |
|---------------------+-------+-------+--------+---------+---------+---------|
| qsort()             | 0.033 | 0.132 |  0.609 |   2.251 |   8.882 |  30.578 |
+---------------------+-------+-------+--------+---------+---------+---------+

```

### ESP8266

* NodeMCU 1.0 clone, 80MHz ESP8266
* Arduino IDE 1.8.13
* ESP8266 Boards 2.7.4

```
+---------------------+------------------------+---------+---------+---------+
|            \      N |    10 |    30 |    100 |     300 |    1000 |    3000 |
| Function    \       |       |       |        |         |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| bubbleSort()        | 0.022 | 0.181 |  2.116 |  19.118 | 213.942 |         |
| insertionSort()     | 0.011 | 0.037 |  0.361 |   3.222 |  34.652 |         |
| selectionSort()     | 0.017 | 0.092 |  0.961 |   8.500 |  93.980 |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| shellSortClassic()  | 0.010 | 0.039 |  0.215 |   0.878 |   3.610 |  13.789 |
| shellSortKnuth()    | 0.010 | 0.036 |  0.172 |   0.690 |   3.021 |  11.306 |
| shellSortTokuda()   | 0.009 | 0.040 |  0.188 |   0.756 |   3.227 |  11.678 |
|---------------------+-------+-------+--------+---------+---------+---------|
| combSort13()        | 0.013 | 0.048 |  0.219 |   0.846 |   3.712 |  13.695 |
| combSort13m()       | 0.013 | 0.051 |  0.223 |   0.840 |   3.625 |  13.102 |
| combSort133()       | 0.010 | 0.040 |  0.208 |   0.796 |   3.571 |  12.430 |
| combSort133m()      | 0.010 | 0.045 |  0.206 |   0.793 |   3.466 |  12.387 |
|---------------------+-------+-------+--------+---------+---------+---------|
| quickSortMiddle()   | 0.013 | 0.045 |  0.185 |   0.650 |   2.518 |   8.303 |
| quickSortMedian()   | 0.016 | 0.052 |  0.203 |   0.684 |   2.574 |   8.470 |
| quickSortMdnSwppd() | 0.012 | 0.039 |  0.160 |   0.560 |   2.128 |   7.129 |
|---------------------+-------+-------+--------+---------+---------+---------|
| qsort()             | 0.027 | 0.092 |  0.417 |   1.515 |   6.009 |  20.782 |
+---------------------+-------+-------+--------+---------+---------+---------+

```

### ESP32

* ESP32-01 Dev Board, 240 MHz Tensilica LX6
* Arduino IDE 1.8.13
* ESP32 Boards 1.0.6

```
+---------------------+------------------------+---------+---------+---------+
|            \      N |    10 |    30 |    100 |     300 |    1000 |    3000 |
| Function    \       |       |       |        |         |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| bubbleSort()        | 0.006 | 0.057 |  0.626 |   5.670 |  64.461 |         |
| insertionSort()     | 0.004 | 0.017 |  0.156 |   1.285 |  14.827 |         |
| selectionSort()     | 0.006 | 0.030 |  0.304 |   2.671 |  29.458 |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| shellSortClassic()  | 0.004 | 0.014 |  0.076 |   0.304 |   1.277 |   4.899 |
| shellSortKnuth()    | 0.004 | 0.013 |  0.062 |   0.245 |   1.102 |   4.205 |
| shellSortTokuda()   | 0.004 | 0.014 |  0.064 |   0.261 |   1.114 |   4.029 |
|---------------------+-------+-------+--------+---------+---------+---------|
| combSort13()        | 0.005 | 0.019 |  0.095 |   0.376 |   1.704 |   6.222 |
| combSort13m()       | 0.006 | 0.020 |  0.098 |   0.381 |   1.640 |   5.996 |
| combSort133()       | 0.005 | 0.019 |  0.093 |   0.368 |   1.649 |   5.668 |
| combSort133m()      | 0.005 | 0.021 |  0.095 |   0.360 |   1.588 |   5.684 |
|---------------------+-------+-------+--------+---------+---------+---------|
| quickSortMiddle()   | 0.005 | 0.017 |  0.071 |   0.252 |   0.987 |   3.427 |
| quickSortMedian()   | 0.006 | 0.018 |  0.073 |   0.255 |   0.969 |   3.263 |
| quickSortMdnSwppd() | 0.004 | 0.014 |  0.059 |   0.209 |   0.815 |   2.804 |
|---------------------+-------+-------+--------+---------+---------+---------|
| qsort()             | 0.008 | 0.033 |  0.144 |   0.522 |   2.131 |   7.245 |
+---------------------+-------+-------+--------+---------+---------+---------+

```

### Teensy 3.2

* 96 MHz ARM Cortex-M4
* Arduino IDE 1.8.13
* Teensyduino 1.53
* Compiler options: "Faster"

```
+---------------------+------------------------+---------+---------+---------+
|            \      N |    10 |    30 |    100 |     300 |    1000 |    3000 |
| Function    \       |       |       |        |         |         |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| bubbleSort()        | 0.009 | 0.083 |  0.945 |   9.217 | 102.362 |         |
| insertionSort()     | 0.005 | 0.040 |  0.399 |   3.290 |  36.427 |         |
| selectionSort()     | 0.009 | 0.065 |  0.653 |   5.715 |  62.889 |         |
|---------------------+-------+-------+--------+---------+---------+---------|
| shellSortClassic()  | 0.007 | 0.029 |  0.159 |   0.648 |   2.694 |  10.372 |
| shellSortKnuth()    | 0.006 | 0.027 |  0.131 |   0.526 |   2.345 |   8.972 |
| shellSortTokuda()   | 0.007 | 0.028 |  0.135 |   0.539 |   2.312 |   8.365 |
|---------------------+-------+-------+--------+---------+---------+---------|
| combSort13()        | 0.008 | 0.037 |  0.182 |   0.720 |   3.234 |  11.948 |
| combSort13m()       | 0.008 | 0.038 |  0.182 |   0.721 |   3.146 |  11.470 |
| combSort133()       | 0.008 | 0.035 |  0.175 |   0.693 |   3.142 |  10.841 |
| combSort133m()      | 0.008 | 0.038 |  0.179 |   0.700 |   3.037 |  10.841 |
|---------------------+-------+-------+--------+---------+---------+---------|
| quickSortMiddle()   | 0.007 | 0.026 |  0.108 |   0.399 |   1.543 |   5.324 |
| quickSortMedian()   | 0.009 | 0.030 |  0.122 |   0.414 |   1.598 |   5.353 |
| quickSortMdnSwppd() | 0.007 | 0.025 |  0.103 |   0.357 |   1.403 |   4.750 |
|---------------------+-------+-------+--------+---------+---------+---------|
| qsort()             | 0.017 | 0.071 |  0.317 |   1.154 |   4.579 |  15.830 |
+---------------------+-------+-------+--------+---------+---------+---------+

```

